# ----------------------------------------------------------------------------
# Copyright (c) 2017-2020, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import skbio
import qiime2
import numpy as np
import scipy


def distance_matrix(metadata: qiime2.NumericMetadataColumn)\
        -> skbio.DistanceMatrix:
    if metadata.has_missing_values():
        missing = metadata.get_ids(where_values_missing=True)
        raise ValueError(
            "Encountered missing value(s) in the metadata column. Computing "
            "a distance matrix from missing values is not supported. IDs with "
            "missing values: %s" % ', '.join(sorted(missing)))

    # This code is derived from @jairideout's scikit-bio cookbook recipe,
    # "Exploring Microbial Community Diversity"
    # https://github.com/biocore/scikit-bio-cookbook
    series = metadata.to_series()
    distances = scipy.spatial.distance.pdist(
        series.values[:, np.newaxis], metric='euclidean')
    return skbio.DistanceMatrix(distances, ids=series.index)



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2020, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest

import pandas as pd
import numpy as np
import skbio
import qiime2

from q2_metadata import distance_matrix


class DistanceMatrixTests(unittest.TestCase):
    def test_int_column(self):
        md = qiime2.NumericMetadataColumn(
            pd.Series([1, 2, 3], name='number',
                      index=pd.Index(['sample1', 'sample2', 'sample3'],
                                     name='id'))
        )
        exp = skbio.DistanceMatrix([[0, 1, 2],
                                    [1, 0, 1],
                                    [2, 1, 0]],
                                   ids=['sample1', 'sample2', 'sample3'])

        obs = distance_matrix(md)

        self.assertEqual(exp, obs)

    def test_float_column(self):
        md = qiime2.NumericMetadataColumn(
            pd.Series([1.5, 2.0, 3.0], name='number',
                      index=pd.Index(['sample1', 'sample2', 'sample3'],
                                     name='id'))
        )
        exp = skbio.DistanceMatrix([[0.0, 0.5, 1.5],
                                    [0.5, 0.0, 1.0],
                                    [1.5, 1.0, 0.0]],
                                   ids=['sample1', 'sample2', 'sample3'])
        obs = distance_matrix(md)

        self.assertEqual(exp, obs)

    def test_one_sample(self):
        md = qiime2.NumericMetadataColumn(
            pd.Series([1.5], name='number',
                      index=pd.Index(['sample1'],
                                     name='id'))
        )
        exp = skbio.DistanceMatrix([[0.0]],
                                   ids=['sample1'])

        obs = distance_matrix(md)

        self.assertEqual(exp, obs)

    def test_missing_values(self):
        md = qiime2.NumericMetadataColumn(
            pd.Series([1.0, 2.0, np.nan, 4.0], name='number',
                      index=pd.Index(['sample1', 'sample2', 'sample3',
                                      'sample4'],
                                     name='id'))
        )

        with self.assertRaisesRegex(ValueError, 'missing values'):
            distance_matrix(md)


if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2020, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import pkg_resources
import shutil

import pandas as pd

import qiime2
import q2templates


TEMPLATES = pkg_resources.resource_filename('q2_metadata', 'templates')


def tabulate(output_dir: str, input: qiime2.Metadata,
             page_size: int = 100) -> None:
    if page_size < 1:
        raise ValueError('Cannot render less than one record per page.')

    df = input.to_dataframe()
    df_columns = pd.MultiIndex.from_tuples(
        [(n, t.type) for n, t in input.columns.items()],
        names=['column header', 'type'])
    df.columns = df_columns
    df.reset_index(inplace=True)
    table = df.to_json(orient='split')
    index = os.path.join(TEMPLATES, 'tabulate', 'index.html')
    q2templates.render(index, output_dir,
                       context={'table': table, 'page_size': page_size})

    input.save(os.path.join(output_dir, 'metadata.tsv'))

    js = os.path.join(TEMPLATES, 'tabulate', 'datatables.min.js')
    os.mkdir(os.path.join(output_dir, 'js'))
    shutil.copy(js, os.path.join(output_dir, 'js', 'datatables.min.js'))

    css = os.path.join(TEMPLATES, 'tabulate', 'datatables.min.css')
    os.mkdir(os.path.join(output_dir, 'css'))
    shutil.copy(css, os.path.join(output_dir, 'css', 'datatables.min.css'))



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2020, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from ._tabulate import tabulate
from ._distance import distance_matrix
from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

__all__ = ['tabulate', 'distance_matrix']



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2020, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from unittest import TestCase, main

from q2_metadata.plugin_setup import plugin


class PluginTests(TestCase):
    def test_plugin(self):
        self.assertEqual(plugin.name, 'metadata')


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2020, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import qiime2.plugin
from qiime2.plugin import MetadataColumn, Numeric

import q2_metadata

from q2_metadata import tabulate, distance_matrix
from q2_types.distance_matrix import DistanceMatrix


plugin = qiime2.plugin.Plugin(
    name='metadata',
    version=q2_metadata.__version__,
    website='https://github.com/qiime2/q2-metadata',
    package='q2_metadata',
    user_support_text=None,
    citation_text=None,
    description=('This QIIME 2 plugin provides functionality for working with '
                 'and visualizing Metadata.'),
    short_description='Plugin for working with Metadata.'
)

plugin.methods.register_function(
    function=distance_matrix,
    inputs={},
    parameters={'metadata': MetadataColumn[Numeric]},
    parameter_descriptions={'metadata': 'Numeric metadata column to compute '
                                        'pairwise Euclidean distances from'},
    outputs=[('distance_matrix', DistanceMatrix)],
    name='Create a distance matrix from a numeric Metadata column',
    description='Create a distance matrix from a numeric metadata column. '
                'The Euclidean distance is computed between each pair of '
                'samples or features in the column.\n\n'
                'Tip: the distance matrix produced by this method can be used '
                'as input to the Mantel test available in `q2-diversity`.'
)

plugin.visualizers.register_function(
    function=tabulate,
    inputs={},
    parameters={
        'input': qiime2.plugin.Metadata,
        'page_size': qiime2.plugin.Int,
    },
    parameter_descriptions={
        'input': 'The metadata to tabulate.',
        'page_size': 'The maximum number of Metadata records to display '
                     'per page',
    },
    name='Interactively explore Metadata in an HTML table',
    description='Generate a tabular view of Metadata. The output '
                'visualization supports interactive filtering, sorting, and '
                'exporting to common file formats.',
)



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2020, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
from unittest import TestCase, main
import tempfile

import pandas as pd
import qiime2

from q2_metadata import tabulate


class TabulateTests(TestCase):
    def test_valid_metadata(self):
        index = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data = ['1.0', '2.0', '3.0']
        md = qiime2.Metadata(pd.DataFrame({'foo': data}, index=index,
                                          dtype=object))

        with tempfile.TemporaryDirectory() as output_dir:
            tabulate(output_dir, md)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))

            viz = open(index_fp).read()

            self.assertTrue('pageLength: 100' in viz)
            self.assertTrue('"columns":[["id",""],["foo","categorical"]]'
                            in viz)
            self.assertTrue(all(i in viz for i in index))
            self.assertTrue(all(val in viz for val in data))

    def test_valid_metadata_many_columns(self):
        index = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data = [['1.0', 'lorem', 'peanut'],
                ['2.0', 'ipsum', 'the'],
                ['3.0', 'emrakul', 'dog']]
        md = qiime2.Metadata(pd.DataFrame(data, index=index, dtype=object,
                                          columns=['foo', 'bar', 'baz']))

        with tempfile.TemporaryDirectory() as output_dir:
            tabulate(output_dir, md)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))

            viz = open(index_fp).read()

            self.assertTrue('pageLength: 100' in viz)
            self.assertTrue('"columns":[["id",""],["foo","categorical"],'
                            '["bar","categorical"],["baz","categorical"]]'
                            in viz)
            self.assertTrue(all(i in viz for i in index))
            self.assertTrue(all(v in viz for row in data for v in row))

    def test_multiple_dtypes(self):
        index = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data = [[1.0, 'lorem'], [2.0, 'ipsum'], [3.0, 'emrakul']]
        md = qiime2.Metadata(pd.DataFrame(data, index=index,
                                          columns=['foo', 'bar']))

        with tempfile.TemporaryDirectory() as output_dir:
            tabulate(output_dir, md)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))

            viz = open(index_fp).read()
            self.assertTrue('pageLength: 100' in viz)
            self.assertTrue('"columns":[["id",""],["foo","numeric"],'
                            '["bar","categorical"]]' in viz)
            self.assertTrue(all(i in viz for i in index))
            self.assertTrue(all(str(v) in viz for row in data for v in row))

    def test_pagination(self):
        index = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        values = ['1.0', '2.0', '3.0']
        md = qiime2.Metadata(pd.DataFrame({'value': values}, index=index))

        with tempfile.TemporaryDirectory() as output_dir:
            tabulate(output_dir, md, page_size=2)
            index_fp = os.path.join(output_dir, 'index.html')
            self.assertTrue(os.path.exists(index_fp))

            viz = open(index_fp).read()

            self.assertTrue('pageLength: 2' in viz)

    def test_invalid_pagination(self):
        index = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        values = ['1.0', '2.0', '3.0']
        md = qiime2.Metadata(pd.DataFrame({'value': values}, index=index))

        with tempfile.TemporaryDirectory() as output_dir:
            with self.assertRaisesRegex(ValueError, 'less than one'):
                tabulate(output_dir, md, -1)


if __name__ == "__main__":
    main()
